from django.contrib.auth.mixins import PermissionRequiredMixin
from django.shortcuts import get_object_or_404, render
from django.views.generic import View

from utilities.views import ObjectDeleteView, ObjectEditView, ObjectListView

from np_autodiscovery import filters, forms, tables
from np_autodiscovery.models import DiscoveryRequest


#
# DiscoveryRequests
#

class DiscoveryRequestListView(PermissionRequiredMixin, ObjectListView):
    permission_required = 'np_autodiscovery.view_discoveryrequest'
    queryset = DiscoveryRequest.objects.all()
    filter = filters.DiscoveryRequestFilter
    filter_form = forms.DiscoveryRequestFilterForm
    table = tables.DiscoveryRequestTable
    template_name = 'np_autodiscovery/discoveryrequest_list.html'


class DiscoveryRequestView(PermissionRequiredMixin, View):
    permission_required = 'np_autodiscovery.view_discoveryrequest'

    def get(self, request, pk):

        discoveryrequest = get_object_or_404(DiscoveryRequest, pk=pk)

        return render(request, 'np_autodiscovery/discoveryrequest.html', {
            'discoveryrequest': discoveryrequest,
        })


class DiscoveryRequestCreateView(PermissionRequiredMixin, ObjectEditView):
    permission_required = 'np_autodiscovery.add_discoveryrequest'
    model = DiscoveryRequest
    model_form = forms.DiscoveryRequestForm
    template_name = 'np_autodiscovery/discoveryrequest_edit.html'
    default_return_url = 'np_autodiscovery:discoveryrequest_list'


class DiscoveryRequestDeleteView(PermissionRequiredMixin, ObjectDeleteView):
    permission_required = 'np_autodiscovery.delete_discoveryrequest'
    model = DiscoveryRequest
    default_return_url = 'np_autodiscovery:discoveryrequest_list'


import django_tables2 as tables

from utilities.tables import BaseTable, BooleanColumn

from np_autodiscovery.models import DiscoveryRequest


STATUS_LABEL = """
<span class="label label-{{ record.get_status_class }}">{{ record.get_status_display }}</span>
"""


class DiscoveryRequestTable(BaseTable):
    prefix = tables.LinkColumn()
    update_existing = BooleanColumn()
    status = tables.TemplateColumn(template_code=STATUS_LABEL, verbose_name='Status')

    class Meta(BaseTable.Meta):
        model = DiscoveryRequest
        fields = ['prefix', 'created', 'update_existing', 'platform', 'status', 'site', 'device_role']



from django.urls import path

from extras.views import ObjectChangeLogView

from np_autodiscovery import views
from np_autodiscovery.models import DiscoveryRequest


app_name = 'np_autodiscovery'
urlpatterns = [

    path('discovery-requests/', views.DiscoveryRequestListView.as_view(), name='discoveryrequest_list'),
    path('discovery-requests/<int:pk>/', views.DiscoveryRequestView.as_view(), name='discoveryrequest'),
    path('discovery-requests/<int:pk>/delete/', views.DiscoveryRequestDeleteView.as_view(), name='discoveryrequest_delete'),
    path('discovery-requests/<int:pk>/changelog/', ObjectChangeLogView.as_view(), name='discoveryrequest_changelog', kwargs={'model': DiscoveryRequest}),
    path('discovery-requests/add/', views.DiscoveryRequestCreateView.as_view(), name='discoveryrequest_add'),

]



# Generated by Django 2.2.3 on 2019-07-10 20:21

from django.db import migrations, models
import django.db.models.deletion
import ipam.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('dcim', '0073_interface_form_factor_to_type'),
    ]

    operations = [
        migrations.CreateModel(
            name='DiscoveryRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('prefix', ipam.fields.IPNetworkField()),
                ('update_existing', models.BooleanField(default=False)),
                ('status', models.PositiveSmallIntegerField(default=1)),
                ('job', models.CharField(blank=True, max_length=50, null=True)),
                ('device_role', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='dcim.DeviceRole')),
                ('platform', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='dcim.Platform')),
                ('site', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='dcim.Site')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='DiscoveryResult',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False)),
                ('address', ipam.fields.IPAddressField()),
                ('status', models.PositiveSmallIntegerField(default=0)),
                ('device', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='dcim.Device')),
                ('discovery_request', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='results', to='np_autodiscovery.DiscoveryRequest')),
            ],
        ),
    ]



from django.apps import AppConfig


class NPAutoDiscoveryConfig(AppConfig):
    name = 'np_autodiscovery'
    verbose_name = 'Auto Discovery'
    netbox_min_version = '2.6.0'
    netbox_max_version = '2.7.0'
    required_configuration_settings = []
    default_configuration_settings = {}
    nav_links = [
        {
            'primary': {
                'name': 'Discovery Requests',
                'view': 'discoveryrequest_list',
                'permission': 'view_discoveryrequest'
            },
            'add': {
                'view': 'discoveryrequest_add',
                'permission': 'add_discoveryrequest'
            },
            #'import': {
            #    'view': 'hello_world',
            #}
        }
    ]



from django.db import models
from django.urls import reverse
from django_rq import get_queue
from rq import cancel_job

from dcim.models import Device, DeviceRole, Platform, Site
from ipam.fields import IPAddressField, IPNetworkField
from utilities.models import ChangeLoggedModel

from np_autodiscovery.constants import *


class DiscoveryRequest(ChangeLoggedModel):
    prefix = IPNetworkField()
    update_existing = models.BooleanField(
        default=False
    )
    status = models.PositiveSmallIntegerField(
        choices=REQUEST_STATUS_CHOICES,
        default=REQUEST_STATUS_PENDING
    )
    platform = models.ForeignKey(
        to=Platform,
        related_name='+',
        on_delete=models.CASCADE
    )
    job = models.CharField(
        max_length=50,
        null=True,
        blank=True
    )
    site = models.ForeignKey(
        to=Site,
        related_name='+',
        on_delete=models.CASCADE
    )
    device_role = models.ForeignKey(
        to=DeviceRole,
        related_name='+',
        on_delete=models.CASCADE
    )

    def __str__(self):
        return '{} {}'.format(self.prefix, self.created)

    def get_absolute_url(self):
        return reverse('np_autodiscovery:discoveryrequest', args=[self.pk])

    def get_status_class(self):
        return REQUEST_STATUS_CLASSES[self.status]

    def save(self, *args, **kwargs):
        if not self.pk:
            new = True
        else:
            new = False

        super().save(*args, **kwargs)

        if new:
            discovery_queue = get_queue('default')
            job = discovery_queue.enqueue(
                "np_autodiscovery.worker.discovery_job",
                self
            )
            self.job = str(job.key)
            self.save()

    def delete(self, *args, **kwargs):
        if self.job:
            try:
                cancel_job(self.job)
            except Exception:
                pass
        super().delete(*args, **kwargs)


class DiscoveryResult(models.Model):
    discovery_request = models.ForeignKey(
        to=DiscoveryRequest,
        related_name='results',
        on_delete=models.CASCADE,
    )
    device = models.ForeignKey(
        to=Device,
        related_name='+',
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    address = IPAddressField()
    status = models.PositiveSmallIntegerField(
        choices=RESULT_STATUS_CHOICES,
        default=RESULT_STATUS_TRYING
    )

    def __str__(self):
        return '{} {}'.format(self.address, self.status)

    def get_absolute_url(self):
        return reverse('np_autodiscovery:discoveryresult', args=[self.pk])

    def get_status_class(self):
        return RESULT_STATUS_CLASSES[self.status]



from django import forms

from dcim.models import DeviceRole, Platform, Site
from utilities.forms import (
    add_blank_choice, APISelect, APISelectMultiple, BootstrapMixin, FilterChoiceField, StaticSelect2
)

from np_autodiscovery.constants import REQUEST_NAPALM_DRIVER_CHOICES, REQUEST_STATUS_CHOICES
from np_autodiscovery.models import DiscoveryRequest


#
# DiscoveryRequest
#

class DiscoveryRequestFilterForm(BootstrapMixin, forms.Form):
    model = DiscoveryRequest
    field_order = ['q', 'status', 'prefix', 'update_existing', 'platform', 'site']
    q = forms.CharField(
        required=False,
        label='Search'
    )
    status = forms.ChoiceField(
        choices=add_blank_choice(REQUEST_STATUS_CHOICES),
        required=False,
        widget=StaticSelect2()
    )
    platform = FilterChoiceField(
        queryset=Platform.objects.all(),
        to_field_name='slug',
        widget=APISelectMultiple(
            api_url="/api/dcim/platforms/",
            value_field="slug",
        )
    )
    site = FilterChoiceField(
        queryset=Site.objects.all(),
        to_field_name='slug',
        widget=APISelectMultiple(
            api_url="/api/dcim/sites/",
            value_field="slug",
        )
    )
    device_role = FilterChoiceField(
        queryset=DeviceRole.objects.all(),
        to_field_name='slug',
        widget=APISelectMultiple(
            api_url="/api/dcim/device-roles/",
            value_field="slug",
        )
    )


class DiscoveryRequestForm(BootstrapMixin, forms.ModelForm):

    class Meta:
        model = DiscoveryRequest
        fields = [
            'prefix', 'update_existing', 'platform', 'site', 'device_role'
        ]
        widgets = {
            'platform': APISelect(
                api_url="/api/dcim/platforms/"
            ),
            'site': APISelect(
                api_url="/api/dcim/sites/"
            ),
            'device_role': APISelect(
                api_url="/api/dcim/device-roles/"
            ),
        }
        help_texts = {
            'prefix': "CIDR preix to scan for discovery",
            'site': "Site in which to look for and create new devices",
            'update_existing': "Update existing devices if they already exist in NetBox",
            'device_role': "Device role to use for new and updated devices",
            'platform': "Device platform which contains the applicable NAPALM driver"
        }



import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='np_autodiscovery',
    version='1.0.2',
    packages=find_packages(),
    include_package_data=True,
    license='Apache 2.0',
    description='NetBox plugin which allows for auto discovery of existing devices in a network.',
    long_description=README,
    url='https://github.com/lampwins/np_autodiscovery/',
    author='John Anderson',
    author_email='lampwins@gmail.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: System :: Networking',
    ],
    install_requires=[
        'napalm',
        'django_rq',
    ]
)



import django_filters
import netaddr
from django.db.models import Q

from dcim.models import DeviceRole, Platform, Site

from np_autodiscovery.constants import REQUEST_STATUS_CHOICES
from np_autodiscovery.models import DiscoveryRequest


class DiscoveryRequestFilter(django_filters.FilterSet):
    q = django_filters.CharFilter(
        method='search',
        label='Search',
    )
    status = django_filters.MultipleChoiceFilter(
        choices=REQUEST_STATUS_CHOICES,
        null_value=None
    )
    prefix = django_filters.CharFilter(
        method='filter_prefix',
        label='Prefix',
    )
    platform_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Platform.objects.all(),
        label='Platform (ID)',
    )
    platform = django_filters.ModelMultipleChoiceFilter(
        field_name='platform__slug',
        queryset=Platform.objects.all(),
        to_field_name='slug',
        label='Platform (slug)',
    )
    site_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Site.objects.all(),
        label='Site (ID)',
    )
    site = django_filters.ModelMultipleChoiceFilter(
        field_name='site__slug',
        queryset=Site.objects.all(),
        to_field_name='slug',
        label='Site (slug)',
    )
    device_role_id = django_filters.ModelMultipleChoiceFilter(
        queryset=DeviceRole.objects.all(),
        label='Device role (ID)',
    )
    device_role = django_filters.ModelMultipleChoiceFilter(
        field_name='device_role__slug',
        queryset=DeviceRole.objects.all(),
        to_field_name='slug',
        label='Device role (slug)',
    )

    class Meta:
        model = DiscoveryRequest
        fields = [
            'id', 'prefix', 'update_existing', 'status',
        ]

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        qs_filter = (
            Q(prefix__icontains=value)
        )
        return queryset.filter(qs_filter)

    def filter_prefix(self, queryset, name, value):
        if not value.strip():
            return queryset
        try:
            query = str(netaddr.IPNetwork(value).cidr)
            return queryset.filter(prefix=query)
        except ValidationError:
            return queryset.none()



import jnpr
import napalm
import netaddr
import socket
from contextlib import closing
from django_rq import job
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from napalm.base.exceptions import (
    ConnectAuthError, ConnectionClosedException, ConnectionException, ConnectTimeoutError, ModuleImportError
)

from dcim.models import Device, DeviceType, Manufacturer

from np_autodiscovery.constants import *
from np_autodiscovery.models import DiscoveryResult


def _get_or_create_device_type(model, manufacturer, vendor):
    device_type = DeviceType.objects.filter(model__iexact=model).first()
    if not device_type:
        # Create a new device type

        # Create a new manufacture if one does not already exist
        if not manufacturer:
            manufacturer = Manufacturer.objects.filter(name__iexact=vendor).first()
            if not manufacturer:
                manufacturer = Manufacturer.objects.create(name=vendor)

        device_type = DeviceType.objects.create(model=model, manufacturer=manufacturer)

    return device_type


@job('default')
def discovery_job(discovery_request):
    """
    Given a DiscoveryRequest object, scan the prefix for devices and take the specified action
    """
    discovery_request.status = REQUEST_STATUS_RUNNING
    discovery_request.save()

    # Validate the configured driver
    try:
        driver = napalm.get_network_driver(discovery_request.platform.napalm_driver)
    except ModuleImportError:
        # No driver is defined on the platform or the driver is invalid
        discovery_request.status = REQUEST_STATUS_FAILED
        discovery_request.save()
        raise ImproperlyConfigured(
            'Platform {} does not have a napalm driver defined or has an invalid driver defined'.format(
                discovery_request.platform
            )
        )

    # Iterate over the prefix
    for address in discovery_request.prefix.iter_hosts():
        discovery_result = DiscoveryResult.objects.create(
            address=address,
            discovery_request=discovery_request
        )

        ip_address = str(address)
        optional_args = settings.NAPALM_ARGS.copy()
        if discovery_request.platform.napalm_args is not None:
            optional_args.update(discovery_request.platform.napalm_args)
        d = driver(
            hostname=ip_address,
            username=settings.NAPALM_USERNAME,
            password=settings.NAPALM_PASSWORD,
            timeout=settings.NAPALM_TIMEOUT,
            optional_args=optional_args
        )

        # Try to connect to the address
        try:
            # Simple open port check first if the port attribute is accessible
            if hasattr(d, 'port'):
                with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
                    sock.settimeout(4)
                    if sock.connect_ex((ip_address, int(d.port))):
                        raise ConnectionException("Port {} not open".format(d.port))

            # Now open the actual connection
            d.open()
        except ConnectAuthError:
            # Login failed
            discovery_result.status = RESULT_STATUS_FAILED_LOGIN
        except jnpr.junos.exception.ConnectError:
            # Connection error
            discovery_result.status = RESULT_STATUS_FAILED_CONNECTION_ERROR
        except (ConnectionException, ConnectTimeoutError, jnpr.junos.exception.ConnectTimeoutError):
            # Unreachable
            discovery_result.status = RESULT_STATUS_UNREACHABLE
        except (ConnectionClosedException, jnpr.junos.exception.ConnectRefusedError):
            # Connection closed
            discovery_result.status = RESULT_STATUS_FAILED_CONNECTION_CLOSED
        except Exception as e:
            # Some other failure
            discovery_result.status = RESULT_STATUS_FAILED
        else:

            # Connection open, now try to get facts
            try:
                device_facts = d.get_facts()
                device_environmental = d.get_environment()
                device_interfaces = d.get_interfaces()
            except Exception:
                # Parse error
                discovery_result.status = RESULT_STATUS_FAILED_PARSING
            else:
                # Everything went well, so now we do something with the results
                d.close()

                device = Device.objects.filter(name=device_facts['hostname'], site=discovery_request.site).first()
                if device:
                    if discovery_request.update_existing:
                        # update
                        discovery_result.status = RESULT_STATUS_COMPLETED_UPDATED_DEVICE

                        device_type = _get_or_create_device_type(
                            device_facts['model'],
                            discovery_request.platform.manufacturer,
                            device_facts['vendor']
                        )

                        device.device_type = device_type
                        device.serial = device_facts['serial_number']
                        device.platform = discovery_request.platform
                        device.device_role = discovery_request.device_role
                        device.save()
                    else:
                        # no update
                        discovery_result.status = RESULT_STATUS_COMPLETED_NO_UPDATED_DEVICE
                else:
                    # new device
                    discovery_result.status = RESULT_STATUS_COMPLETED_NEW_DEVICE

                    device_type = _get_or_create_device_type(
                        device_facts['model'],
                        discovery_request.platform.manufacturer,
                        device_facts['vendor']
                    )

                    device = Device.objects.create(
                        name=device_facts['hostname'],
                        device_type=device_type,
                        serial=device_facts['serial_number'],
                        site=discovery_request.site,
                        platform=discovery_request.platform,
                        device_role=discovery_request.device_role
                    )

                discovery_result.device = device

        discovery_result.save()

    discovery_request.status = REQUEST_STATUS_COMPLETE
    discovery_request.save()
    return True



REQUEST_STATUS_PENDING = 1
REQUEST_STATUS_RUNNING = 2
REQUEST_STATUS_COMPLETE = 3
REQUEST_STATUS_FAILED = 4
REQUEST_STATUS_CHOICES = [
    [REQUEST_STATUS_PENDING, 'Pending'],
    [REQUEST_STATUS_RUNNING, 'Running'],
    [REQUEST_STATUS_COMPLETE, 'Complete'],
    [REQUEST_STATUS_FAILED, 'Failed'],
]

REQUEST_STATUS_CLASSES = {
    1: 'default',
    2: 'warning',
    3: 'success',
    4: 'danger',
}


REQUEST_NAPALM_DRIVER_JUNOS = 1
REQUEST_NAPALM_DRIVER_CHOICES = [
    [REQUEST_NAPALM_DRIVER_JUNOS, 'Junos']
]

RESULT_STATUS_TRYING = 0
RESULT_STATUS_UNREACHABLE = 1
RESULT_STATUS_COMPLETED_NEW_DEVICE = 2
RESULT_STATUS_COMPLETED_UPDATED_DEVICE = 3
RESULT_STATUS_COMPLETED_NO_UPDATED_DEVICE = 4
RESULT_STATUS_FAILED_LOGIN = 5
RESULT_STATUS_FAILED_PARSING = 6
RESULT_STATUS_FAILED = 7
RESULT_STATUS_FAILED_CONNECTION_CLOSED = 8
RESULT_STATUS_FAILED_EXISTING_CONFLICT = 9
RESULT_STATUS_FAILED_CONNECTION_ERROR = 10
RESULT_STATUS_CHOICES = [
    [RESULT_STATUS_TRYING, 'Trying'],
    [RESULT_STATUS_UNREACHABLE, 'Unreachable'],
    [RESULT_STATUS_COMPLETED_NEW_DEVICE, 'Complete - New Device'],
    [RESULT_STATUS_COMPLETED_UPDATED_DEVICE, 'Complete - Updated Device'],
    [RESULT_STATUS_COMPLETED_NO_UPDATED_DEVICE, 'Complete - No Update to Device'],
    [RESULT_STATUS_FAILED_LOGIN, 'Failed - Login'],
    [RESULT_STATUS_FAILED_PARSING, 'Failed - Parsing'],
    [RESULT_STATUS_FAILED, 'Failed'],
    [RESULT_STATUS_FAILED_CONNECTION_CLOSED, 'Failed - Connection Closed'],
    [RESULT_STATUS_FAILED_EXISTING_CONFLICT, 'Failed - Existing Device(s) Conflict'],
    [RESULT_STATUS_FAILED_CONNECTION_ERROR, 'Failed - Connection Error'],
]

RESULT_STATUS_CLASSES = {
    0: 'info',
    1: 'warning',
    2: 'success',
    3: 'success',
    4: 'success',
    5: 'danger',
    6: 'danger',
    7: 'danger',
    8: 'danger',
    9: 'danger',
    10: 'danger',
}
